/*******************************************************************************
!#                              INTEL CONFIDENTIAL
!#
!#  Copyright 2010-2013 Intel Corporation All Rights Reserved.
!#
!#  The  source code  contained  or  described herein and all documents related
!#  related to the source code  ("Material")  are  owned by  Intel  Corporation
!#  or  its  suppliers  or  licensors.  Title  to  the  Material  remains  with
!#  Intel Corporation or its suppliers and licensors. The  Material may contain
!#  trade secrets  and  proprietary   and   confidential information  of  Intel
!#  or its  suppliers  and  licensors, and is protected by  worldwide copyright
!#  and  trade secret laws  and  treaty  provisions.  No  part of the  Material
!#  may used,   copied,  reproduced,  modified,  published,  uploaded,  posted,
!#  transmitted,  distributed,  or disclosed in any way without  Intel's  prior
!#  express written permission. No license under any  patent,  copyright, trade
!#  secret or other intellectual property right is granted to or conferred upon
!#  you  by  disclosure  or  delivery of the  Materials,  either  expressly, by
!#  implication, inducement,  estoppel or otherwise.  Any  license  under  such
!#  intellectual  property  rights must be express and  approved  by  Intel  in
!#  writing.
!#
!#  Unless otherwise  agreed  by  Intel in writing, you may not remove or alter
!#  this notice or any other notice embedded in  Materials  by Intel or Intel's
!#  suppliers or licensors in any way.
!#
!#*****************************************************************************/
#include <stdio.h>
#include <math.h>
#include <malloc.h>
#ifdef __linux__
#include <sys/time.h>
#else
#include <windows.h>
#endif
#include "multiply.h"

// Timer implementation
#ifdef __linux__
static struct timeval tstart, tstop;
#else
static FILETIME tstart, tstop;
#endif
void start_timer()
{
#ifdef __linux__
    gettimeofday(&tstart, NULL);
#else
    GetSystemTimeAsFileTime(&tstart);
#endif
}

double stop_timer()
{
#ifdef __linux__
    gettimeofday(&tstop, NULL);
    return (tstop.tv_sec + tstop.tv_usec / 1000000.0) -
           (tstart.tv_sec + tstart.tv_usec / 1000000.0);
#else
    GetSystemTimeAsFileTime(&tstop);
    return (double)(tstop.dwLowDateTime - tstart.dwLowDateTime) / 10000000;
#endif
}

// Initialize matrix
void init_matrix(int row, int col, FTYPE off, FTYPE  a[][COLWIDTH])
{
    int i,j;

    for (i = 0; i < row; i++) {
        for (j = 0; j < col; j++) {
            a[i][j] = fmod(i * j + off, 10.0);
        }
    }

    if (COLBUF > 0) // optional padding
        for  (i = 0; i < row; i++)
            for (j = col; j < COLWIDTH; j++)
                a[i][j] = 0.0;
}

// Initialize vector
void init_vector(int length, FTYPE off, FTYPE a[])
{
    int i;

    for (i = 0; i < length; i++)
        a[i] = fmod(i + off, 10.0);

    if (COLBUF > 0) // optional padding
        for  (i = length; i < COLWIDTH; i++)
            a[i] = 0.0;
}

// Print the sum of the first "length" elements of "vec"
// (needed to verify correctness of different versions)
void printsum(int length, FTYPE vec[]) {
    int i;
    FTYPE sum=0.0;

    for (i=0; i<length; i++) sum += vec[i];

    printf("Sum of result = %f\n", sum);
}

// For multiply.c; defined here to simulate what happens if loop strides are not
// known during compilation time.
unsigned int inc_i = 1;
unsigned int inc_j = 1;

int main(int argc, char **argv)
{
    int i;
    double duration = 0.0;

    __declspec(align(32)) FTYPE a[ROW][COLWIDTH];
    __declspec(align(32)) FTYPE b[ROW] = { 0.0 };
    __declspec(align(32)) FTYPE x[COLWIDTH];

    printf("ROW: %d COL: %d\n", ROW, COL);

    // Initialize matrix & vector
    init_matrix(ROW, COL, 1.0, a);
    init_vector(COL, 3.0, x);

    // Do the measurement
    start_timer();
    for (i = 0; i < REPEATNTIMES; i++) {
        matvec(ROW, COLWIDTH, a, b, x);
    }
    duration = stop_timer();

    // Print results
    printf("Elapsed time = %lf seconds\n", duration);
    printf("GigaFlops = %f\n", (((double)REPEATNTIMES *
                                 (double)ROW * (double)COLWIDTH * 2.0) /
                                 duration) / 1000000000.0);
    printsum(ROW, b);

    return 0;
}
