/*
// Copyright 2003-2012 Intel Corporation. All Rights Reserved.
// 
// The source code contained or described herein and all documents related 
// to the source code ("Material") are owned by Intel Corporation or its
// suppliers or licensors.  Title to the Material remains with Intel Corporation
// or its suppliers and licensors.  The Material is protected by worldwide
// copyright and trade secret laws and treaty provisions.  No part of the
// Material may be used, copied, reproduced, modified, published, uploaded,
// posted, transmitted, distributed, or disclosed in any way without Intel's
// prior express written permission.
// 
// No license under any patent, copyright, trade secret or other intellectual
// property right is granted to or conferred upon you by disclosure or delivery
// of the Materials, either expressly, by implication, inducement, estoppel
// or otherwise.  Any license under such intellectual property rights must
// be express and approved by Intel in writing.

*/


#include<stdio.h>
#include<stdlib.h>
#include"offload.h"
#include"time.h"
#include <sys/time.h>
#include<omp.h>

#define SIZE (1024*1024*64)
#define BLOCK (1024*1024*1)
#define NUM_BLOCK (SIZE/BLOCK)
#define ITERS (NUM_BLOCK*2) 
#define PHI_DEV 0
#define ALPHA 1.3f
#define PASSES 10
#define NUM_THREADS 240

#define ALLOC alloc_if(1)
#define FREE free_if(1)
#define REUSE alloc_if(0)
#define RETAIN free_if(0)

//#define DEBUG
#define VALIDATE

__declspec(target(mic)) static float *ina0, *inb0, *out0, *ina1, *inb1, *out1;

float *ref0,*ref1;

void do_async(void)
{
	int block0_transfer_number=0, block1_transfer_number=0;
        //Send array 0 to the coprocessor
        #pragma offload_transfer target(mic:PHI_DEV)\
                in(ina0:length(BLOCK) REUSE RETAIN)\
                in(inb0:length(BLOCK) REUSE RETAIN)\
		signal(&ina0)
        {
        }
	block0_transfer_number++;

        
	for(int j=0;j<ITERS;j++)
	{
	if((j%2)==0)
	{
		#ifdef DEBUG
		printf("ITER %d OFFLOAD 1 IN1:TRASNFER BLOCK %d\n\n",j,block1_transfer_number);fflush(0);
		#endif
		//Start the asynchronous transfer for array 1 and send signal "a1" when complete
 	        #pragma offload_transfer target(mic:PHI_DEV)\
         		in(ina1[block1_transfer_number*BLOCK:BLOCK] : REUSE RETAIN)\
               		in(inb1[block1_transfer_number*BLOCK:BLOCK] : REUSE RETAIN)\
			if(j!=(ITERS-1))\
                	signal(&ina1)
        	{
        	}
		block1_transfer_number++;
		
		#ifdef DEBUG
                printf("ITER %d OFFLOAD 2 IN0: COMPUTE BLOCK %d\n\n",j,block0_transfer_number-1);fflush(0);
                #endif
        	//Start simultaneous compute while the transfer for array 1 is occuring
        	#pragma offload target(mic:PHI_DEV)\
                	nocopy(ina0:length(0) REUSE RETAIN)\
                	nocopy(inb0:length(0) REUSE RETAIN)\
                	out(out0[(block0_transfer_number-1)*BLOCK:BLOCK] : REUSE RETAIN)\
			wait(&ina0)
        	{
                	//Perform SAXPY operation
                	#pragma omp parallel for firstprivate(ina0,inb0,out0)
                	for(int i=0;i<BLOCK;i++)
                	{
                        	out0[(block0_transfer_number-1)*BLOCK+i]=ALPHA*ina0[(block0_transfer_number-1)*BLOCK+i]+inb0[(block0_transfer_number-1)*BLOCK+i];
                	}
        	}
	}
	else 
	{
		#ifdef DEBUG
                printf("ITER %d OFFLOAD 1 IN0: TRANSFER BLOCK %d\n\n",j,block0_transfer_number);fflush(0);
                #endif
        	//Start the asynchronous transfer for out0
        	#pragma offload_transfer target(mic:PHI_DEV)\
			in(ina0[block0_transfer_number*BLOCK:BLOCK] : REUSE RETAIN)\
			in(inb0[block0_transfer_number*BLOCK:BLOCK] : REUSE RETAIN)\
			if(j!=(ITERS-1))\
                	signal(&ina0)
        	{
        	}
		block0_transfer_number++;


		#ifdef DEBUG
                printf("ITER %d OFFLOAD 2 IN1: COMPUTE BLOCK %d\n\n",j,block1_transfer_number-1);fflush(0);
                #endif
        	//Start simultaneous compute while the transfer for out0 is occuring once you receive signal "array1" and transfer out1 back to host once done.
       		#pragma offload target(mic:PHI_DEV)\
                	nocopy(ina1:length(0) REUSE RETAIN)\
                	nocopy(inb1:length(0) REUSE RETAIN)\
                	out(out1[(block1_transfer_number-1)*BLOCK:BLOCK] : REUSE RETAIN)\
                	wait(&ina1)
        	{
                	//Perform SAXPY operation
                	#pragma omp parallel for firstprivate(ina1,inb1,out1)
                	for(int i=0;i<BLOCK;i++)
                	{
                        	out1[(block1_transfer_number-1)*BLOCK+i]=ALPHA*ina1[(block1_transfer_number-1)*BLOCK+i]+inb1[(block1_transfer_number-1)*BLOCK+i];
                	}
        	}
	}
	}
}	

void do_offload(void)
{
	
	//Offload 1
	#pragma offload target(mic:PHI_DEV)\
		in(ina0:length(SIZE) REUSE RETAIN)\
                in(inb0:length(SIZE) REUSE RETAIN)\
                out(out0:length(SIZE) REUSE RETAIN)
	{
		//Perform SAXPY operation 
		#pragma omp parallel for firstprivate(ina0,inb0,out0)
		for(int i=0;i<SIZE;i++)
		{
			out0[i]=ALPHA*ina0[i]+inb0[i];
		}
	}

	//Offload 2
	#pragma offload target(mic:PHI_DEV)\
		in(ina1:length(SIZE) REUSE RETAIN)\
                in(inb1:length(SIZE) REUSE RETAIN)\
                out(out1:length(SIZE) REUSE RETAIN)
	{
		//Perform SAXPY operation
		#pragma omp parallel for firstprivate(ina1,inb1,out1)
                for(int i=0;i<SIZE;i++)
                {
                        out1[i]=ALPHA*ina1[i]+inb1[i];
                }
	}
}

int main()
{
	struct timeval start,end;
	double runtime;

	//Allocate arrays on host 
	ina0=(float*)malloc(sizeof(float)*SIZE);
	inb0=(float*)malloc(sizeof(float)*SIZE);
	ina1=(float*)malloc(sizeof(float)*SIZE);
	inb1=(float*)malloc(sizeof(float)*SIZE);
	out0=(float*)malloc(sizeof(float)*SIZE);
	out1=(float*)malloc(sizeof(float)*SIZE);
	ref0=(float*)malloc(sizeof(float)*SIZE);
	ref1=(float*)malloc(sizeof(float)*SIZE);	

	//Initialize host arrays
	srand(time(NULL));
	for(int i=0;i<SIZE;i++)
	{
		ina0[i]=(float)(rand()%100);
		inb0[i]=(float)(rand()%100);
                out0[i]=0.0f;;
                ina1[i]=(float)(rand()%100);
                inb1[i]=(float)(rand()%100);
                out1[i]=0.0f;	
		ref0[i]=ALPHA*ina0[i]+inb0[i];
		ref1[i]=ALPHA*ina1[i]+inb1[i];
	}

	//Allocate the arrays on coprocessor
	#pragma offload target(mic:PHI_DEV)\
		nocopy(ina0:length(SIZE) ALLOC RETAIN)\
		nocopy(inb0:length(SIZE) ALLOC RETAIN)\
		nocopy(out0:length(SIZE) ALLOC RETAIN)\
		nocopy(ina1:length(SIZE) ALLOC RETAIN)\
                nocopy(inb1:length(SIZE) ALLOC RETAIN)\
                nocopy(out1:length(SIZE) ALLOC RETAIN)
	{
		//Set number of threads on the coprocessor
		omp_set_num_threads(NUM_THREADS);
	}

	gettimeofday(&start,NULL);
	for(int i=0;i<PASSES;i++)
	{
	do_offload();
	}
	gettimeofday(&end, NULL);
        runtime=(float) (end.tv_sec - start.tv_sec)+ (float) (end.tv_usec - start.tv_usec) * 1e-6;
        printf("Offload Runtime: %.3f\n",runtime/PASSES);
	
	//Validate the output 
	#ifdef VALIDATE
	for(int i=0;i<SIZE;i++)
	{
		if(abs(ref0[i]-out0[i])>1e-6)
		{
			printf("ERROR:%d\n",i);
		}
	}
	#endif

	gettimeofday(&start,NULL);
        for(int i=0;i<PASSES;i++)
        {
        do_async();
        }
        gettimeofday(&end, NULL);
        runtime=(float) (end.tv_sec - start.tv_sec)+ (float) (end.tv_usec - start.tv_usec) * 1e-6;
        printf("Async Runtime: %.3f\n",runtime/PASSES);

	//Validate the output
	#ifdef VALIDATE
	for(int i=0;i<SIZE;i++)
        {
                if((abs(ref0[i]-out0[i])>1e-3))
                {
                        printf("ERROR:%d\n",i);break;
                }
        }
	#endif

	//Free the arrays on the coprocessor
	#pragma offload target(mic:PHI_DEV)\
                nocopy(ina0:length(SIZE) REUSE FREE)\
                nocopy(inb0:length(SIZE) REUSE FREE)\
                nocopy(out0:length(SIZE) REUSE FREE)\
                nocopy(ina1:length(SIZE) REUSE FREE)\
                nocopy(inb1:length(SIZE) REUSE FREE)\
                nocopy(out1:length(SIZE) REUSE FREE)
        {
        }

	return 0;
}
