// Copyright 2003-2012 Intel Corporation. All Rights Reserved.
// 
// The source code contained or described herein and all documents related 
// to the source code ("Material") are owned by Intel Corporation or its
// suppliers or licensors.  Title to the Material remains with Intel Corporation
// or its suppliers and licensors.  The Material is protected by worldwide
// copyright and trade secret laws and treaty provisions.  No part of the
// Material may be used, copied, reproduced, modified, published, uploaded,
// posted, transmitted, distributed, or disclosed in any way without Intel's
// prior express written permission.
// 
// No license under any patent, copyright, trade secret or other intellectual
// property right is granted to or conferred upon you by disclosure or delivery
// of the Materials, either expressly, by implication, inducement, estoppel
// or otherwise.  Any license under such intellectual property rights must
// be express and approved by Intel in writing.


#include <stdio.h>
#include "timer.h"
#include <stdlib.h>
#include "malloc.h"
#include "offload.h"
#include <math.h>
#include <mkl.h>
#include <mkl_vsl.h>

#define max(a, b) (((a)>(b))?(a):(b))
#define TIMESTEPS 1000
#define PATHS 1000000

/* program to solve Monte Carlo problem */
main()
{
  float secs = 0.0f;
  double T, r, sigma, rho, alpha, dt, con1, con2, xmu, var;
  double Y1, Y2, S1, S2, rvec0[TIMESTEPS], rvec1[TIMESTEPS];
  double payoff, sum0, sum1, ave, sdv;
  int    paths, timeSteps, mm, nn;
  int seed = 1;
  VSLStreamStatePtr stream0, stream1;

  Timer myTimer; /*timer*/
  
  myTimer.start();
  paths = PATHS;
  timeSteps = TIMESTEPS; 

  printf("Monte Carlo simulation of %d paths with %d timesteps\n", paths, timeSteps);

  /* set various constants */
  T     = 1.0;
  r     = 0.05;
  sigma = 0.1;
  rho   = 0.5;
  alpha = sqrt(1.0-rho*rho);

  dt    = 1.0/timeSteps;

  con1  = 1.0 + r*dt;
  con2  = sqrt(dt)*sigma;

  xmu = 0.0;
  var = 1.0;
  payoff = 0.951229424500714; // exp(-r*T);

  /* initialise VSL random number generator */
  vslNewStream(&stream0, VSL_BRNG_MCG59, seed);
  vslCopyStream(&stream1, stream0);  // Create a second instance of the same randome stream
  vslLeapfrogStream( stream0, 0, 2); // The master stream will have 2 partitions.  0 takes
  vslLeapfrogStream( stream1, 1, 2); // the first, 1 the second

  /* loop over paths */
  sum0 = 0.0;
  sum1 = 0.0;
  for (mm=0; mm<paths; mm++) 
  {

    /* calculate random vector */
    vdRngGaussian(VSL_METHOD_DGAUSSIAN_BOXMULLER2,
                  stream0, timeSteps, rvec0, xmu, var);
    vdRngGaussian(VSL_METHOD_DGAUSSIAN_BOXMULLER2,
                  stream1, timeSteps, rvec1, xmu, var);

    /* calculate path  */
    S1 = 1.0;
    S2 = 1.0;

    for (nn=0; nn<timeSteps; nn++) 
    {
      Y1 = rvec0[nn];
      Y2 = rho*Y1 + alpha*rvec1[nn];

      S1 = S1*(con1 + con2*Y1);
      S2 = S2*(con1 + con2*Y2);
    }

    if ( fabs(S1-1.0)<0.1 && fabs(S2-1.0)<0.1 ) 
    {
      sum0 = sum0 + payoff;
      sum1 = sum1 + payoff*payoff;
    }
  }

  /* calculate average and standard deviation */
  ave = sum0/paths;
  sdv = sqrt((sum1/paths - ave*ave)/paths);
  vslDeleteStream(&stream0);
  vslDeleteStream(&stream1);
  secs = myTimer.end();
  printf("Time Elapsed: %f secs \n", secs);
  printf("Average %f, Standard deviation of error %f\n", ave, sdv);
}
