// Copyright 2003-2012 Intel Corporation. All Rights Reserved.
// 
// The source code contained or described herein and all documents related 
// to the source code ("Material") are owned by Intel Corporation or its
// suppliers or licensors.  Title to the Material remains with Intel Corporation
// or its suppliers and licensors.  The Material is protected by worldwide
// copyright and trade secret laws and treaty provisions.  No part of the
// Material may be used, copied, reproduced, modified, published, uploaded,
// posted, transmitted, distributed, or disclosed in any way without Intel's
// prior express written permission.
// 
// No license under any patent, copyright, trade secret or other intellectual
// property right is granted to or conferred upon you by disclosure or delivery
// of the Materials, either expressly, by implication, inducement, estoppel
// or otherwise.  Any license under such intellectual property rights must
// be express and approved by Intel in writing.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "offload.h"
#include "unistd.h"
#include "omp.h"

#define PHIDEV 0 

/// --------------- Globals
__attribute__((target(mic))) omp_lock_t output_lock;

__attribute__((target(mic))) int done = 0;

__attribute__((target(mic))) char buffer[1024];

/// ---------------- Offloaded functions
__attribute__((target(mic))) void workloadinit(void)
{
   omp_init_lock(&output_lock);
}

__attribute__((target(mic))) void mainworkload(char *inputstr)
{
   // Initialize the buffer
   omp_set_lock(&output_lock);

#ifdef __MIC__
   printf("PHI mainworkload - %s\n", inputstr);
   fflush(NULL);
#else
   printf("HOST mainworkload - %s\n", inputstr);
   fflush(NULL);
#endif

   strcpy(buffer, inputstr);

   omp_unset_lock(&output_lock);

   for (int i = 0; i < 10; i++) {
      // Hold lock until buffer ready to copy
      omp_set_lock(&output_lock);
      sleep(1);
      strcpy(buffer, inputstr);
      strcat(buffer, " micbuffer");
      for (int j = 0; j < i; j++) strcat(buffer, ".");
      omp_unset_lock(&output_lock);
   }
#ifdef __MIC__
   printf("PHI mainworkload completed\n");fflush(0);
   fflush(NULL);
#else
   printf("HOST mainworkload completed \n");fflush(0);
   fflush(NULL);
#endif

   // Signal all done
   done = 1;
}

__attribute__((target(mic))) int getresults(char *outputstr)
{
  if (done != 1)
  {
      // Hold lock while copying buffer
      omp_set_lock(&output_lock);
      strcpy(outputstr, buffer);
      omp_unset_lock(&output_lock);
   }
   return done;
}

void work_1(char *instring) {
	// Send data to main workload, which will run for a while
#pragma offload target(mic:PHIDEV) in(instring) inout(buffer)
	mainworkload(instring);        
}

int work_2(char *outstring) {
	int status = 0;
    #pragma offload target(mic:PHIDEV) out(status,outstring)
    status = getresults(outstring);
	return status;
}

//////////////////////////////////////////////////////////////////////////////////
main(int argc, char* argv[])
{
    char instring[1024];
    char outstring[1024];
    strcpy(instring, "hostinput -");

    #pragma offload target(mic:PHIDEV)
    {
        // Make sure our lock is initialized ahead of use
        workloadinit();
    }

#pragma omp parallel 
#pragma omp single
   {
       #pragma omp task
       {
			work_1(instring);
		  }
       #pragma omp task
       {
	      // Get buffers from the kernel as they are ready
	      int status = 0;
          do
          {
			status = work_2(outstring);
		  if (status == 0)
		printf("HOST - long-running kernel returns: %s\n", outstring);
             else
                printf("HOST - status = %d\n", status);
             fflush(NULL);
          }
	  while( status == 0);
       } // omp task
   } // parallel omp single and parallel
    
   // End of calls to test functions
   printf("\nHOST - Test completed\n");
}
