/*******************************************************************************
!#                              INTEL CONFIDENTIAL
!#
!#  Copyright 2010-2013 Intel Corporation All Rights Reserved.
!#
!#  The  source code  contained  or  described herein and all documents related
!#  related to the source code  ("Material")  are  owned by  Intel  Corporation
!#  or  its  suppliers  or  licensors.  Title  to  the  Material  remains  with
!#  Intel Corporation or its suppliers and licensors. The  Material may contain
!#  trade secrets  and  proprietary   and   confidential information  of  Intel
!#  or its  suppliers  and  licensors, and is protected by  worldwide copyright
!#  and  trade secret laws  and  treaty  provisions.  No  part of the  Material
!#  may used,   copied,  reproduced,  modified,  published,  uploaded,  posted,
!#  transmitted,  distributed,  or disclosed in any way without  Intel's  prior
!#  express written permission. No license under any  patent,  copyright, trade
!#  secret or other intellectual property right is granted to or conferred upon
!#  you  by  disclosure  or  delivery of the  Materials,  either  expressly, by
!#  implication, inducement,  estoppel or otherwise.  Any  license  under  such
!#  intellectual  property  rights must be express and  approved  by  Intel  in
!#  writing.
!#
!#  Unless otherwise  agreed  by  Intel in writing, you may not remove or alter
!#  this notice or any other notice embedded in  Materials  by Intel or Intel's
!#  suppliers or licensors in any way.
!#
!#*****************************************************************************/
#include <stdio.h>
#ifdef __linux__
#include <sys/time.h>
#else
#include <windows.h>
#endif

#define RUNS 1000000

extern float simd(unsigned int stride, unsigned int size, float *a);

// Timer implementation
#ifdef __linux__
static struct timeval tstart, tstop;
#else
static FILETIME tstart, tstop;
#endif
void start_timer()
{
#ifdef __linux__
    gettimeofday(&tstart, NULL);
#else
    GetSystemTimeAsFileTime(&tstart);
#endif
}

double stop_timer()
{
#ifdef __linux__
    gettimeofday(&tstop, NULL);
    return (tstop.tv_sec + tstop.tv_usec / 1000000.0) -
           (tstart.tv_sec + tstart.tv_usec / 1000000.0);
#else
    GetSystemTimeAsFileTime(&tstop);
    return (double)(tstop.dwLowDateTime - tstart.dwLowDateTime) / 10000000;
#endif
}

int main(int argc, char **argv)
{
    int i;
    float sum = 0.0;
    double duration = 0.0;

#define N 10000
    float a[N];

    // Initialize array
    for(i = 0; i < N; i++) a[i] = i;

#define OFFSET 4 // Safe to have vector length up to 4
    start_timer();
    for(i = 0; i < RUNS; i++) sum += simd(OFFSET, N, a);
    duration = stop_timer();

    // Print results
    printf("Elapsed time = %lf seconds\n", duration);
    printf("Sum: %f\n", sum);

    return 0;
}
